#!/usr/bin/env python
# encoding: utf-8

"""Project definition module."""


from os import sep
from os.path import (basename, dirname, exists, isabs, isdir, join, relpath, 
  splitext)
from weakref import WeakValueDictionary
from zipfile import ZipFile
from .util import AzkabanError, temppath
import logging
import sys


logger = logging.getLogger(__name__)


class Project(object):

  """Azkaban project.

  :param name: name of the project
  :param register: add project to registry. setting this to false will make it
    invisible to the CLI
  :param root: optional path to a root file or directory used to enable adding
    files with relative paths (typically used with root=__file__)

  """

  _registry = WeakValueDictionary()

  def __init__(self, name, root=None, register=True):
    self.name = name
    self.root = root if not root or isdir(root) else dirname(root)
    if register:
      self._registry[name] = self
    self._jobs = {}
    self._files = {}

  def __str__(self):
    return self.name

  @property
  def files(self):
    """Returns a list of files that will be included in the project archive.

    This property should not be used to add files. Use `add_files` instead.

    """
    return [relpath(e) for e in self._files]

  @property
  def jobs(self):
    """Returns a dictionary with each job options.

    This property should not be used to add jobs. Use `add_job` instead.

    """
    return dict(
      (name, job.options)
      for name, job in self._jobs.items()
    )

  def add_file(self, path, archive_path=None):
    """Include a file in the project archive.

    :param path: absolute path to file
    :param archive_path: path to file in archive (defaults to same as `path`)

    This method requires the path to be absolute to avoid having files in the
    archive with lower level destinations than the base root directory.

    """
    logger.debug('adding file %r as %r', path, archive_path or path)
    if not isabs(path):
      if self.root:
        path = join(self.root, path)
      else:
        raise AzkabanError(
          'Relative path not allowed without specifying a project root: %r.'
          % (path, )
        )
    if path in self._files:
      if self._files[path] != archive_path:
        raise AzkabanError('Inconsistent duplicate file: %r.' % (path, ))
    else:
      if not exists(path):
        raise AzkabanError('File not found: %r.' % (path, ))
      self._files[path] = archive_path

  def add_job(self, name, job):
    """Include a job in the project.

    :param name: name assigned to job (must be unique)
    :param job: `Job` subclass

    This method triggers the `on_add` method on the added job (passing the
    project and name as arguments). The handler will be called right after the
    job is added.

    """
    logger.debug('adding job %r', name)
    if name in self._jobs:
      raise AzkabanError('Duplicate job name: %r.' % (name, ))
    else:
      self._jobs[name] = job
      job.on_add(self, name)

  def merge_into(self, project):
    """Merge one project with another.

    :param project: project to merge with this project

    This method does an in place merge of the current project with another.
    The merged project will maintain the current project's name.
    """
    logger.debug('merging into project %r', project.name)
    for name, job in self._jobs.items():
      project.add_job(name, job)
    for path, archive_path in self._files.items():
      project.add_file(path, archive_path)

  def build(self, path, overwrite=False):
    """Create the project archive.

    :param path: destination path
    :param overwrite: don't throw an error if a file already exists at `path`

    Triggers the `on_build` method on each job inside the project (passing
    itself and the job's name as two argument). This method will be called
    right before the job file is generated.

    """
    logger.debug('building project')
    # not using a with statement for compatibility with older python versions
    if exists(path) and not overwrite:
      raise AzkabanError('Path %r already exists.' % (path, ))
    if not (len(self._jobs) or len(self._files)):
      raise AzkabanError('Building empty project.')
    writer = ZipFile(path, 'w')
    try:
      for name, job in self._jobs.items():
        job.on_build(self, name)
        with temppath() as fpath:
          job.build(fpath)
          writer.write(fpath, '%s.job' % (name, ))
      for fpath, apath in self._files.items():
        writer.write(fpath, apath)
    finally:
      writer.close()

  @classmethod
  def load_from_script(cls, script, name=None):
    """Get project from script.

    :param script: string representing a python module
    :param name: project name

    If `name` is unspecified:
    * if a single project is found by loading `script` that project is returned.
    * in any other case, an error is thrown.

    """
    sys.path.append(dirname(script))
    module = splitext(basename(script.rstrip(sep)))[0]
    try:
      __import__(module)
    except ImportError:
      raise AzkabanError('Unable to import script %r.' % (script, ))
    else:
      if name:
        try:
          return cls._registry[name]
        except KeyError:
          raise AzkabanError(
            'Unable to find project with name %r in script %r.\n'
            'Available projects: %r.'
            % (name, script, ', '.join(cls._registry.keys()))
          )
      else:
        if len(cls._registry) == 1:
          return cls._registry.popitem()[1]
        elif not cls._registry:
          raise AzkabanError('No project found in %r.' % (script, ))
        else:
          raise AzkabanError(
            'Multiple projects found in %r: %s.\n'
            'Disambiguate using --project=%s:project_name.'
            % (', '.join(cls._registry.keys()), script, script)
          )



#!/usr/bin/env python
# encoding: utf-8

"""Utility module."""


from ConfigParser import (NoOptionError, NoSectionError, ParsingError,
  RawConfigParser)
from contextlib import contextmanager
from functools import wraps
from os import close, remove
from os.path import exists, expanduser
from requests import ConnectionError
from requests.exceptions import MissingSchema
from tempfile import mkstemp
import requests as rq
import sys


class AzkabanError(Exception):

  """Base error class."""

  def __init__(self, message, *args):
    super(AzkabanError, self).__init__(message % args or ())


class Config(object):

  """Configuration class."""

  def __init__(self, path=expanduser('~/.azkabanrc')):
    if not exists(path):
      raise AzkabanError('No configuration file found at %r.', path)
    self.parser = RawConfigParser()
    self.path = path
    try:
      self.parser.read(self.path)
    except ParsingError:
      raise AzkabanError('Invalid configuration file %r.', path)

  def save(self):
    """Save configuration parser back to file."""
    with open(self.path, 'w') as writer:
      self.parser.write(writer)

  def get_default_option(self, command, name):
    """Get default option value for a command.

    :param command: command the option should be looked up for
    :param name: name of the option

    """
    try:
      return self.parser.get(command, 'default.%s' % (name, ))
    except (NoOptionError, NoSectionError):
      raise AzkabanError(
        'In order to omit the `--%(name)s` option, '
        'a default %(name)s must be defined in %(path)r.\n'
        'This is done by adding a `default.%(name)s` option in the '
        '`%(command)s` section.'
        % {'command': command, 'name': name, 'path': self.path}
      )


@contextmanager
def temppath():
  """Create a temporary filepath.

  Usage::

    with temppath() as path:
      # do stuff

  Any file corresponding to the path will be automatically deleted afterwards.

  """
  (desc, path) = mkstemp()
  close(desc)
  remove(path)
  try:
    yield path
  finally:
    if exists(path):
      remove(path)

def catch(*error_classes):
  """Returns a decorator that catches errors and prints messages to stderr.

  :param *error_classes: error classes

  Also exits with status 1 if any errors are caught.

  """
  def decorator(func):
    """Decorator."""
    @wraps(func)
    def wrapper(*args, **kwargs):
      """Wrapper. Finally."""
      try:
        return func(*args, **kwargs)
      except error_classes as err:
        sys.stderr.write('%s\n' % (err, ))
        sys.exit(1)
    return wrapper
  return decorator

def flatten(dct, sep='.'):
  """Flatten a nested dictionary.

  :param dct: dictionary to flatten.
  :param sep: separator used when concatenating keys.

  """
  def _flatten(dct, prefix=''):
    """Inner recursive function."""
    items = []
    for key, value in dct.items():
      new_prefix = '%s%s%s' % (prefix, sep, key) if prefix else key
      if isinstance(value, dict):
        items.extend(_flatten(value, new_prefix).items())
      else:
        items.append((new_prefix, value))
    return dict(items)
  return _flatten(dct)

def human_readable(size):
  """Transform size from bytes to human readable format (kB, MB, ...).

  :param size: size in bytes

  """
  for suffix in ['bytes', 'kB', 'MB', 'GB', 'TB']:
    if size < 1024.0:
      return '%3.1f%s' % (size, suffix)
    size /= 1024.0

def azkaban_request(method, url, **kwargs):
  """Make request to azkaban server and catch common errors.

  :param method: get, post, etc.
  :param url: endpoint url
  :param kwargs: arguments forwarded to the request handler

  This function is meant to handle common errors and return a more helpful
  message than the default one.

  """
  try:
    handler = getattr(rq, method.lower())
  except AttributeError:
    raise ValueError('Invalid HTTP method: %r.' % (method, ))
  else:
    try:
      response = handler(url, verify=False, **kwargs)
    except ConnectionError:
      raise AzkabanError('Unable to connect to azkaban at %r.' % (url, ))
    except MissingSchema:
      raise AzkabanError('Invalid azkaban server url: %r.' % (url, ))
    else:
      return response

def extract_json(response):
  """Extract json from Azkaban response, gracefully handling errors.

  :param response: request response object

  """
  try:
    json = response.json()
  except ValueError:
    # no json decoded probably
    raise ValueError('No JSON decoded from response %r' % (response.text, ))
  else:
    if 'error' in json:
      raise AzkabanError(json['error'])
    elif json.get('status') == 'error':
      raise AzkabanError(json['message'])
    else:
      return json



#!/usr/bin/env python
# encoding: utf-8

"""Azkaban session module.

This contains the `Session` class which will be used for all interactions with
a remote Azkaban server.

"""


from ConfigParser import NoOptionError, NoSectionError
from getpass import getpass, getuser
from os.path import exists
from .util import AzkabanError, Config, azkaban_request, extract_json
import logging


logger = logging.getLogger(__name__)


def _parse_url(url):
  """Parse url.

  :param url: http endpoint (including protocol, port and optional user)

  """
  parsed_url = url.rstrip('/').split('@')
  if len(parsed_url) == 1:
    user = getuser()
    url = parsed_url[0]
  elif len(parsed_url) == 2:
    user = parsed_url[0]
    url = parsed_url[1]
  else:
    raise AzkabanError('Malformed url: %r' % (url, ))
  return (user, url)

def _resolve_alias(config, alias):
  """Get url associated with an alias.

  :param alias: alias name

  """
  try:
    return config.parser.get('alias', alias)
  except (NoOptionError, NoSectionError):
    raise AzkabanError('Alias %r not found.', alias)

def _get_session_id(config, url):
  """Retrieve session id associated with url.

  :param url: http endpoint (including protocol, port and optional user)

  """
  try:
    return config.parser.get('session_id', url)
  except NoSectionError:
    config.parser.add_section('session_id')
  except NoOptionError:
    pass


class Session(object):

  """Azkaban session.

  :param url: http endpoint (including protocol, port and optional user)
  :param alias: alias name

  """

  def __init__(self, url=None, alias=None):
    self.config = Config()
    if not url:
      alias = alias or self.config.get_default_option('azkaban', 'alias')
      url = _resolve_alias(self.config, alias)
    self.user, self.url = _parse_url(url)
    self.id = _get_session_id(self.config, str(self).replace(':', '.'))
    logger.debug('session %s instantiated', self)

  def __str__(self):
    return '%s@%s' % (self.user, self.url)

  def _refresh(self, password=None):
    """Refresh session ID.

    :param password: password used to log into Azkaban (only used if no alias
      is provided). can be set to `False` to fail instead of prompting for a
      password.

    Also saves session ID for future use.

    """
    logger.debug('refreshing session')
    password = password or getpass('Azkaban password for %s: ' % (self, ))
    res = extract_json(azkaban_request(
      'POST',
      self.url,
      data={'action': 'login', 'username': self.user, 'password': password},
    ))
    self.id = res['session.id']
    logger.debug('saving session id')
    self.config.parser.set('session_id', str(self).replace(':', '.'), self.id)
    self.config.save()

  def _request(self, method, endpoint, use_cookies=True, attempts=1, **kwargs):
    """Make request to Azkaban using this session.

    :param method: http method
    :param endpoint: server endpoint (e.g. manager)
    :param attempts: if current session ID is invalid, maximum number of
      attempts to refresh it
    :param use_cookies: include session_id in cookies instead of request data
    :param **kwargs: keyword arguments passed to `azkaban.util.azkaban_request`

    If the session expired, will prompt for a password to refresh.

    """
    full_url = '%s/%s' % (self.url, endpoint.lstrip('/'))
    logger.debug('sending request to %r: %r', full_url, kwargs)
    while True:
      if use_cookies:
        kwargs.setdefault('cookies', {})['azkaban.browser.session.id'] = self.id
      else:
        kwargs.setdefault('data', {})['session.id'] = self.id
      res = azkaban_request(method, full_url, **kwargs) if self.id else None
      if res is None or '<!-- /.login -->' in res.text:
        # checking for None because 500 responses evaluate to False
        logger.debug('request failed because of invalid login')
        if attempts > 0:
          self._refresh()
          attempts -= 1
        else:
          raise AzkabanError(
            'Too many unsuccessful login attempts for url %r. Aborting.',
            self.url
          )
      else:
        return res

  def get_execution_status(self, exec_id):
    """Get status of an execution.

    :param exec_id: execution ID

    """
    return extract_json(self._request(
      method='GET',
      endpoint='executor',
      params={
        'execid': exec_id,
        'ajax': 'fetchexecflow',
      },
    ))

  def get_job_logs(self, exec_id, job, offset=0, limit=50000):
    """Get logs from a job execution.

    :param exec_id: execution ID
    :param job: job name
    :param offset: log offset
    :param limit: size of log to download

    """
    return extract_json(self._request(
      method='GET',
      endpoint='executor',
      params={
        'execid': exec_id,
        'jobId': job,
        'ajax': 'fetchExecJobLogs',
        'offset': offset,
        'length': limit,
      },
    ))

  def cancel_execution(self, exec_id):
    """Cancel workflow execution.

    :param exec_id: execution ID

    """
    res = extract_json(self._request(
      method='GET',
      endpoint='executor',
      params={
        'execid': exec_id,
        'ajax': 'cancelFlow',
      },
    ))
    if 'error' in res:
      raise AzkabanError('Execution %s is not running.' % (exec_id, ))
    return res

  def create_project(self, name, description):
    """Create project.

    :param name: project name
    :param description: project description

    """
    return extract_json(self._request(
      method='POST',
      endpoint='manager',
      data={
        'action': 'create',
        'name': name,
        'description': description,
      },
    ))

  def delete_project(self, name):
    """Delete a project on Azkaban.

    :param session: `azkaban.util.Session` object

    """
    res = self._request(
      method='GET',
      endpoint='manager',
      params={
        'project': name,
        'delete': 'true',
      },
    )
    msg = "Project '%s' was successfully deleted" % (name, )
    if not msg in res.text:
      raise AzkabanError('Delete failed. Check permissions and existence.')
    return res

  def run_workflow(self, project, flow, jobs=None, skip=False):
    """Launch a workflow.

    :param project: name of the project
    :param flow: name of the workflow
    :param jobs: name of jobs to run (run entire workflow by default)
    :param skip: don't run if the same workflow is already running

    Note that in order to run a workflow on Azkaban, it must already have been
    uploaded and the corresponding user must have permissions to run.

    """
    if not jobs:
      disabled = '[]'
    else:
      all_names = set(
        n['id']
        for n in self.get_workflow_info(project, flow)['nodes']
      )
      run_names = set(jobs)
      missing_names = run_names - all_names
      if missing_names:
        raise AzkabanError(
          'Jobs not found in flow %r: %s.' %
          (flow, ', '.join(missing_names))
        )
      else:
        disabled = (
          '[%s]'
          % (','.join('"%s"' % (n, ) for n in all_names - run_names), )
        )
    return extract_json(self._request(
      method='POST',
      endpoint='executor',
      use_cookies=False,
      data={
        'ajax': 'executeFlow',
        'project': project,
        'flow': flow,
        'disabled': disabled,
        'concurrentOption': 'skip' if skip else 'concurrent',
      },
    ))

  def upload_project(self, project, path):
    """Upload project archive.

    :param project: project name
    :param path: path to zip archive

    """
    if not exists(path):
      raise AzkabanError('Unable to find archive at %r.' % (path, ))
    return extract_json(self._request(
      method='POST',
      endpoint='manager',
      use_cookies=False,
      data={
        'ajax': 'upload',
        'project': project,
      },
      files={
        'file': ('file.zip', open(path, 'rb'), 'application/zip'),
      },
    ))

  def get_workflow_info(self, project, flow):
    """Get list of jobs corresponding to a workflow.

    :param project: project name
    :param flow: name of flow in project

    """
    raw_res = self._request(
      method='GET',
      endpoint='manager',
      params={
        'ajax': 'fetchflowjobs',
        'project': project,
        'flow': flow,
      },
    )
    try:
      return extract_json(raw_res)
    except ValueError:
      raise AzkabanError('Flow %r not found.' % (flow, ))



#!/usr/bin/env python
# encoding: utf-8

"""AzkabanPig: an extension for pig scripts to Azkaban CLI.

Usage:
  azkabanpig PATH ...
             [-j JAR] ... [-o OPTION] ...
             [-bp PROJECT] [-t TYPE] [-a ALIAS | -u URL]
  azkabanpig -h | --help

Arguments:
  PATH                          Path to pig script. If more than one path is
                                specified, they will be run in order. Note that
                                all the pig jobs will share the same Azkaban
                                options; for more flexibility, you can use the
                                Azkaban CLI directly.

Options:
  -a ALIAS --alias=ALIAS        Cf. `azkaban --help`.
  -b --background               Run job asynchronously. `azkabanpig` will
                                launch the workflow and return.
  -h --help                     Show this message and exit.
  -j JAR --jar=JAR              Path to jar file. It will be available on the
                                class path when the pig script is run, no need
                                to register it inside your scripts.
  -o OPTION --option=OPTION     Azkaban option. Should be of the form
                                key=value. E.g. '-o param.foo=bar' will
                                substitute parameter '$foo' with 'bar' in the
                                pig script.
  -p PROJECT --project=PROJECT  Project name under which to run the pig script.
  -t TYPE --type=TYPE           Pig job type used.
  -u URL --url=URL              Cf. `azkaban --help`.

Examples:
  azkabanpig my_script.pig
  azkabanpig -a foo first_script.pig second_script.pig
  azkabanpig -url http://url.to.azkaban -o param.my_output=bar.dat foo.pig

AzkabanPig returns with exit code 1 if an error occurred and 0 otherwise.

"""

__all__ = ['PigJob']

from docopt import docopt
from os import sep
from os.path import abspath, basename
from sys import stdout
from time import sleep
from ..job import Job
from ..project import Project
from ..session import Session
from ..util import AzkabanError, Config, catch, temppath


class PigJob(Job):

  """Job class corresponding to pig jobs.

  :param path: absolute path to pig script (this script will automatically be
    added to the project archive)
  :param options: cf. `Job`

  """

  #: Job type used (change this to use a custom pig type).
  type = 'pig'

  def __init__(self, path, *options):
    super(PigJob, self).__init__(
      {'type': self.type, 'pig.script': path.lstrip(sep)},
      *options
    )
    self.path = path

  def on_add(self, project, name):
    """This handler adds the corresponding script file to the project."""
    project.add_file(self.path)


class PigProject(Project):

  """Project to run pig jobs from the command line.

  :param name: project name used
  :param paths: paths to pig scripts, these will be run in order
  :param type: pig job type used
  :param jars: jars to include, these will also be added to the classpath
  :param options: options forwarded to pig scripts

  """

  def __init__(self, name, paths, user, type='pig', jars=None, options=None):
    super(PigProject, self).__init__(name, register=False)
    jars = jars or []
    opts = [
      {
        'type': type,
        'user.to.proxy': user,
        'pig.additional.jars': ','.join(abspath(j).lstrip(sep) for j in jars),
      },
      options or {}
    ]
    for jar in jars:
      self.add_file(abspath(jar))
    for path, dep in zip(paths, [None] + paths):
      dep_opts = {'dependencies': basename(dep)} if dep else {}
      self.add_job(basename(path), PigJob(abspath(path), dep_opts, *opts))


def _get_status(session, exec_id):
  """Get status of a PigProject execution, along with currently running job.

  :param session: `azkaban.session.Session`
  :param exec_id: execution ID

  This method is able to simply find which unique job is active because of
  the linear structure of the workflow.

  """
  status = session.get_execution_status(exec_id)
  active = [
    e['id']
    for e in status['nodes']
    if e['status'] == 'RUNNING' or e['status'] == 'FAILED'
  ]
  return {
    'flow_status': status['status'],
    'active_job': active[0] if active else '',
  }

@catch(AzkabanError)
def main():
  """AzkabanPig entry point."""
  args = docopt(__doc__)
  config = Config()
  paths = args['PATH']
  pj = args['--project'] or config.get_default_option('azkabanpig', 'project')
  tpe = args['--type'] or config.get_default_option('azkabanpig', 'type')
  session = Session(args['--url'], args['--alias'])
  try:
    job_options = dict(opt.split('=', 1) for opt in args['--option'])
  except ValueError:
    raise AzkabanError('Invalid options: %r' % (' '.join(args['OPTION']), ))
  project = PigProject(
    name=pj,
    paths=paths,
    user=session.user,
    type=tpe,
    jars=args['--jar'],
    options=job_options,
  )
  with temppath() as tpath:
    project.build(tpath)
    while True:
      try:
        session.upload_project(project, tpath)
      except AzkabanError:
        session.create_project(project, project)
      else:
        break
  res = session.run_workflow(project, basename(paths[-1]))
  exec_id = res['execid']
  if args['--background']:
    if len(paths) == 1:
      stdout.write(
        'Pig job running at %s/executor?execid=%s&job=%s\n'
        % (session.url, exec_id, basename(paths[0]))
      )
    else:
      stdout.write(
        'Pig jobs running at %s/executor?execid=%s\n'
        % (session.url, exec_id)
      )
  else:
    current_job = None
    try:
      while True:
        sleep(5)
        status = _get_status(session, exec_id)
        if status['active_job'] != current_job:
          current_job = status['active_job']
          if current_job:
            offset = 0
            stdout.write('\n[Job %s]\n' % (current_job, ))
        if current_job:
          logs = session.get_job_logs(
            exec_id=exec_id,
            job=current_job,
            offset=offset,
          )
          stdout.write(logs['data'])
          offset += logs['length']
        if status['flow_status'] == 'SUCCEEDED':
          stdout.write('\nExecution succeeded.\n')
          break
        elif status['flow_status'] != 'RUNNING':
          raise AzkabanError('Execution failed.')
    except KeyboardInterrupt:
      choice = raw_input('\nKill execution [yN]? ').lower()
      if choice and choice[0] == 'y':
        stdout.write('Killing... ')
        session.cancel_execution(exec_id)
        stdout.write('Done.\n')
      else:
        stdout.write(
          'Execution still running at %s/executor?execid=%s\n'
          % (session.url, exec_id)
        )

if __name__ == '__main__':
  main()
